#ifndef LIBKERN_INCLUDED

#define LIBKERN_INCLUDED
/*
**  s_KernelUni(const vX, const vY, const avYK)
**
**  Purpose:
**    Fit a kernel through a dataset
**
**  Inputs:
**    vX    1 x iK vector with support points
**    vY    1 x iN vector of data
**
**  Outputs:
**    avYK  1 x iK vector of corresponding y-values for the kernel
**
**  Remark:
**    A Gaussian kernel is applied, with an automatic bandwidth
**    selection based on Silverman's rule. 
**
**  This library is based on code by Ruud Koning, Gary King, Mico Loretan,
**    Geoffrey Shuetrim and Berwin Turlach, and got adapted by
**    Dick van Dijk (in Gauss) and Charles Bos (in Ox)
*/
s_KernelUni(const vX, const vY, const avYK)
{
  decl iT, iqr, ds, dh, ir, vArg, vK;

  iT= columns(vY[0][]);
  iqr= quantiler(vY[0][], .75) - quantiler(vY[0][], .25);
  ds = sqrt(varr(vY[0][]));
  dh= 0.9 * min(ds, (iqr / 1.34)) / iT .^ 0.2;

  avYK[0] = zeros(vX);

  ir= 0;
  if (dh > 0)
    {
      /* Compute density estimates */
      vArg= (vX' - vY[0][]) / dh;
      vK= densn(vArg);

      avYK[0] = meanr(vK)'/dh;
      ir= 1;
    }

  return ir;
}

/*
**  s_KernelMult(const vX, const vY, const avYK)
**
**  Purpose:
**    Fit a kernel through a multivariate dataset
**
**  Inputs:
**    mX    iP x iK matrix with support points
**    mY    iP x iN matrix of observed data points
**
**  Outputs:
**    amYK  1 x iK matrix of corresponding y-values for the kernel
**
**  Remark:
**    A Gaussian kernel is applied, with an automatic bandwidth
**    selection based on Silverman's rule. 
**
**  Based on routine mKernel, by Dick van Dijk.
**
**  Author:
**    Charles Bos
*/
s_KernelMult(const mX, const mY, const amYK)
{
  decl iN, iP, iK, iqr, ds, mh, mih, ddeth, ir, mArg, mK, i, 
       dLogDet, dSignDet;

  iP= rows(mX);
  iK= columns(mX);
  iN= columns(mY);

  if (iP == 1)
    {
      iqr= quantiler(mY[0][], .75) - quantiler(mY[0][], .25);
      ds = sqrt(varr(mY[0][]));
      mh= 0.9 * min(ds, (iqr / 1.34)) / iN .^ 0.2;
      mih= 1/mh;
      ddeth= mh;
    }
  else
    {
      mh= variance(mY');
      mh= choleski(mh).*iN^(-1/(iP+4));
      mih=invert(mh, &dLogDet, &dSignDet);
      ddeth=exp(dLogDet)*dSignDet;
    }

  amYK[0] = zeros(1, iK);

  ir= 0;
  if (!(mh == 0))
    {
      /* Compute density estimates */
      for (i= 0; i < iK; ++i)
        {
          mArg= mih * (mX[][i] - mY);
          mK= densn(mArg);
          mK= prodc(mK) ./ ddeth;

          amYK[0][0][i] = meanr(mK);
        }
      ir= 1;
    }

  return ir;
}
#endif // LIBKERN_INCLUDED
